@@ -1,6 +1,6 @@
 ;;; xr-test.el --- Tests for xr.el                   -*- lexical-binding: t -*-
 
-;; Copyright (C) 2019 Free Software Foundation, Inc.
+;; Copyright (C) 2019-2020 Free Software Foundation, Inc.
 
 ;; Author: Mattias Engdegård <mattiase@acm.org>
 
@@ -185,6 +185,8 @@
                  '(any "z-" digit)))
   (should (equal (xr "[A-[:digit:]]")
                  '(seq (any "A-[" ":dgit") "]")))
+  (should (equal (xr "[^\n]")
+                 'nonl))
   (should-error (xr "[[::]]"))
   (should-error (xr "[[:=:]]"))
   (should-error (xr "[[:letter:]]"))
@@ -340,117 +342,255 @@
   )
 
 (ert-deftest xr-lint ()
-  (should (equal (xr-lint "^a*\\[\\?\\$\\(b\\{3\\}\\|c\\)[^]\\a-d^-]$")
-                 nil))
-  (should (equal (xr-lint "a^b$c")
-                 '((1 . "Unescaped literal `^'")
-                   (3 . "Unescaped literal `$'"))))
-  (should (equal (xr-lint "^**$")
-                 '((1 . "Unescaped literal `*'"))))
-  (should (equal (xr-lint "a[\\\\[]")
-                 '((3 . "Duplicated `\\' inside character alternative"))))
-  (should (equal (xr-lint "\\{\\(+\\|?\\)\\[\\]\\}\\\t")
-                 '((0  . "Escaped non-special character `{'")
-                   (4  . "Unescaped literal `+'")
-                   (7  . "Unescaped literal `?'")
-                   (14 . "Escaped non-special character `}'")
-                   (16 . "Escaped non-special character `\\t'"))))
-  (should (equal (xr-lint "\\}\\w\\a\\b\\%")
-                 '((0 . "Escaped non-special character `}'")
-                   (4 . "Escaped non-special character `a'")
-                   (8 . "Escaped non-special character `%'"))))
-  (should (equal (xr-lint "a?+b+?\\(?:c?\\)*d\\{3\\}+e*?\\{2,5\\}")
-                 '((2  . "Repetition of repetition")
-                   (14 . "Repetition of repetition")
-                   (25 . "Repetition of repetition"))))
-  (should (equal (xr-lint "\\(a*\\)*\\(b+\\)*\\(c*\\)?\\(d+\\)?")
-                 '((6 .  "Repetition of repetition")
-                   (13 . "Repetition of repetition")
-                   (20 . "Repetition of repetition"))))
-  (should (equal (xr-lint "\\(e*\\)\\{3\\}")
-                 '((6 . "Repetition of repetition"))))
-  (should (equal (xr-lint "[]-Qa-fz-t]")
-                 '((1 . "Reversed range `]-Q' matches nothing")
-                   (7 . "Reversed range `z-t' matches nothing"))))
-  (should (equal (xr-lint "[z-a][^z-a]")
-                 nil))
-  (should (equal (xr-lint "[^A-FFGI-LI-Mb-da-eg-ki-ns-t33-7]")
-                 '((5  . "Character `F' included in range `A-F'")
-                   (10 . "Ranges `I-L' and `I-M' overlap")
-                   (16 . "Ranges `a-e' and `b-d' overlap")
-                   (22 . "Ranges `g-k' and `i-n' overlap")
-                   (25 . "Two-character range `s-t'")
-                   (29 . "Character `3' included in range `3-7'"))))
-  (should (equal (xr-lint "[a[:digit:]b[:punct:]c[:digit:]]")
-                 '((22 . "Duplicated character class `[:digit:]'"))))
-  (should (equal (xr-lint "a*\\|b+\\|\\(?:a\\)*")
-                 '((8 . "Duplicated alternative branch"))))
-  (should (equal (xr-lint "a\\{,\\}")
-                 '((1 . "Uncounted repetition"))))
-  (should (equal (xr-lint "a\\{\\}")
-                 '((1 . "Implicit zero repetition"))))
-  (should (equal (xr-lint "[0-9[|]*/]")
-                 '((4 . "Suspect `[' in char alternative"))))
-  (should (equal (xr-lint "[^][-].]")
-                 nil))
-  (should (equal (xr-lint "[0-1]")
-                 nil))
-  (should (equal (xr-lint "[^]-][]-^]")
-                 '((6 . "Two-character range `]-^'"))))
-  (should (equal
-           (xr-lint "[-A-Z][A-Z-][A-Z-a][^-A-Z][]-a][A-Z---.]")
-           '((16 . "Literal `-' not first or last in character alternative"))))
-  (should (equal
-           (xr-lint "\\(?:a*b?\\)*\\(c\\|d\\|\\)+\\(^\\|e\\)*\\(?:\\)*")
-           '((10 . "Repetition of expression matching an empty string")
-             (21 . "Repetition of expression matching an empty string"))))
-  (should (equal (xr-lint "\\'*\\<?\\(?:$\\)+")
-                 '((2 . "Repetition of zero-width assertion")
-                   (5 . "Repetition of zero-width assertion")
-                   (13 . "Repetition of zero-width assertion"))))
-  (should (equal (xr-lint "a.cde*f?g\\|g\\|abcdefg")
-                 '((14 . "Branch matches subset of a previous branch"))))
-  (should (equal (xr-lint "abcd\\|e\\|[aA].[^0-9z]d")
-                 '((9 . "Branch matches superset of a previous branch"))))
-  (should (equal (xr-lint "\\(?:\\(a\\)\\|.\\)\\(?:a\\|\\(.\\)\\)")
-                 '((21 . "Branch matches superset of a previous branch"))))
-  (should (equal (xr-lint ".\\|\n\\|\r")
-                 '((6 . "Branch matches subset of a previous branch"))))
-  (should (equal (xr-lint "[^mM]\\|[^a-zA-Z]")
-                 '((7 . "Branch matches subset of a previous branch"))))
-  (should (equal (xr-lint "[^mM]\\|[^A-LN-Z]")
-                 nil))
-  (should (equal (xr-lint "[ab]\\|[^bcd]")
-                 nil))
-  (should (equal (xr-lint "[ab]\\|[^cd]")
-                 '((6 . "Branch matches superset of a previous branch"))))
-  (should (equal (xr-lint ".\\|[a\n]")
-                 nil))
-  (should (equal (xr-lint ".\\|[[:space:]\r]")
-                 '((3 . "Branch matches subset of a previous branch"))))
-  (should (equal (xr-lint "ab?c+\\|a?b*c*")
-                 '((7 . "Branch matches superset of a previous branch"))))
-  (should (equal (xr-lint "\\(?:[aA]\\|b\\)\\|a")
-                 '((15 . "Branch matches subset of a previous branch"))))
-  (should (equal (xr-lint "\\(?:a\\|b\\)\\|[abc]")
-                 '((12 . "Branch matches superset of a previous branch"))))
-  (should (equal (xr-lint "\\(?:a\\|b\\)\\|\\(?:[abd]\\|[abc]\\)")
-                 '((12 . "Branch matches superset of a previous branch"))))
-  (should (equal (xr-lint "ab\\|abc?")
-                 '((4 . "Branch matches superset of a previous branch"))))
-  (should (equal (xr-lint "abc\\|abcd*e?")
-                 '((5 . "Branch matches superset of a previous branch"))))
-  (should (equal (xr-lint "[ab]+a?,a?[ab]+,[ab]*a*,a*[ab]*")
-                 '((6 . "Repetition subsumed by preceding repetition")
-                   (14 . "Repetition subsumes preceding repetition")
-                   (22 . "Repetition subsumed by preceding repetition")
-                   (30 . "Repetition subsumes preceding repetition"))))
-  (should (equal (xr-lint "[^a]+b*,a?.*,a*a*,a*a+")
-                 '((6 . "Repetition subsumed by preceding repetition")
-                   (11 . "Repetition subsumes preceding repetition")
-                   (16 . "Repetition subsumed by preceding repetition")
-                   (21 . "Repetition subsumes preceding repetition"))))
-  )
+  (let ((text-quoting-style 'grave))
+    (should (equal (xr-lint "^a*\\[\\?\\$\\(b\\{3\\}\\|c\\)[^]\\a-d^-]$")
+                   nil))
+    (should (equal (xr-lint "a^b$c")
+                   '((1 . "Unescaped literal `^'")
+                     (3 . "Unescaped literal `$'"))))
+    (should (equal (xr-lint "^**$")
+                   '((1 . "Unescaped literal `*'"))))
+    (should (equal (xr-lint "a[\\\\[]")
+                   '((3 . "Duplicated `\\' inside character alternative"))))
+    (should (equal (xr-lint "\\{\\(+\\|?\\)\\[\\]\\}\\\t")
+                   '((0  . "Escaped non-special character `{'")
+                     (4  . "Unescaped literal `+'")
+                     (7  . "Unescaped literal `?'")
+                     (14 . "Escaped non-special character `}'")
+                     (16 . "Escaped non-special character `\\t'"))))
+    (should (equal (xr-lint "\\}\\w\\a\\b\\%")
+                   '((0 . "Escaped non-special character `}'")
+                     (4 . "Escaped non-special character `a'")
+                     (8 . "Escaped non-special character `%'"))))
+    (should (equal (xr-lint "a?+b+?\\(?:c?\\)*d\\{3\\}+e*?\\{2,5\\}")
+                   '((2  . "Repetition of repetition")
+                     (14 . "Repetition of repetition")
+                     (25 . "Repetition of repetition"))))
+    (should (equal (xr-lint "\\(a*\\)*\\(b+\\)*\\(c*\\)?\\(d+\\)?")
+                   '((6 .  "Repetition of repetition")
+                     (13 . "Repetition of repetition")
+                     (20 . "Repetition of repetition"))))
+    (should (equal (xr-lint "\\(e*\\)\\{3\\}")
+                   '((6 . "Repetition of repetition"))))
+    (should (equal (xr-lint "[]-Qa-fz-t]")
+                   '((1 . "Reversed range `]-Q' matches nothing")
+                     (7 . "Reversed range `z-t' matches nothing"))))
+    (should (equal (xr-lint "[z-a][^z-a]")
+                   nil))
+    (should (equal (xr-lint "[^A-FFGI-LI-Mb-da-eg-ki-ns-t33-7]")
+                   '((5  . "Character `F' included in range `A-F'")
+                     (10 . "Ranges `I-L' and `I-M' overlap")
+                     (16 . "Ranges `a-e' and `b-d' overlap")
+                     (22 . "Ranges `g-k' and `i-n' overlap")
+                     (25 . "Two-character range `s-t'")
+                     (29 . "Character `3' included in range `3-7'"))))
+    (should (equal (xr-lint "[a[:digit:]b[:punct:]c[:digit:]]")
+                   '((22 . "Duplicated character class `[:digit:]'"))))
+    (should (equal (xr-lint "a*\\|b+\\|\\(?:a\\)*")
+                   '((8 . "Duplicated alternative branch"))))
+    (should (equal (xr-lint "a\\{,\\}")
+                   '((1 . "Uncounted repetition"))))
+    (should (equal (xr-lint "a\\{\\}")
+                   '((1 . "Implicit zero repetition"))))
+    (should (equal (xr-lint "[0-9[|]*/]")
+                   '((4 . "Suspect `[' in char alternative"))))
+    (should (equal (xr-lint "[^][-].]")
+                   nil))
+    (should (equal (xr-lint "[0-1]")
+                   nil))
+    (should (equal (xr-lint "[^]-][]-^]")
+                   '((6 . "Two-character range `]-^'"))))
+    (should (equal
+             (xr-lint "[-A-Z][A-Z-][A-Z-a][^-A-Z][]-a][A-Z---.]")
+             '((16 .
+                "Literal `-' not first or last in character alternative"))))
+    (should (equal
+             (xr-lint "\\(?:a*b?\\)*\\(c\\|d\\|\\)+\\(^\\|e\\)*\\(?:\\)*")
+             '((10 . "Repetition of expression matching an empty string")
+               (21 . "Repetition of expression matching an empty string"))))
+    (should (equal (xr-lint "\\'*\\<?\\(?:$\\)+")
+                   '((2 . "Repetition of zero-width assertion")
+                     (5 . "Repetition of zero-width assertion")
+                     (13 . "Repetition of zero-width assertion"))))
+    ))
+
+(ert-deftest xr-lint-branch-subsumption ()
+  (let ((text-quoting-style 'grave))
+    (should (equal (xr-lint "a.cde*f?g\\|g\\|abcdefg")
+                   '((14 . "Branch matches subset of a previous branch"))))
+    (should (equal (xr-lint "abcd\\|e\\|[aA].[^0-9z]d")
+                   '((9 . "Branch matches superset of a previous branch"))))
+    (should (equal (xr-lint "\\(?:\\(a\\)\\|.\\)\\(?:a\\|\\(.\\)\\)")
+                   '((21 . "Branch matches superset of a previous branch"))))
+    (should (equal (xr-lint ".\\|\n\\|\r")
+                   '((6 . "Branch matches subset of a previous branch"))))
+    (should (equal (xr-lint "[^mM]\\|[^a-zA-Z]")
+                   '((7 . "Branch matches subset of a previous branch"))))
+    (should (equal (xr-lint "[^mM]\\|[^A-LN-Z]")
+                   nil))
+    (should (equal (xr-lint "[ab]\\|[^bcd]")
+                   nil))
+    (should (equal (xr-lint "[ab]\\|[^cd]")
+                   '((6 . "Branch matches superset of a previous branch"))))
+    (should (equal (xr-lint ".\\|[a\n]")
+                   nil))
+    (should (equal (xr-lint "ab?c+\\|a?b*c*")
+                   '((7 . "Branch matches superset of a previous branch"))))
+    (should (equal (xr-lint "\\(?:[aA]\\|b\\)\\|a")
+                   '((15 . "Branch matches subset of a previous branch"))))
+    (should (equal (xr-lint "\\(?:a\\|b\\)\\|[abc]")
+                   '((12 . "Branch matches superset of a previous branch"))))
+    (should (equal (xr-lint "\\(?:a\\|b\\)\\|\\(?:[abd]\\|[abc]\\)")
+                   '((12 . "Branch matches superset of a previous branch"))))
+    (should (equal (xr-lint "ab\\|abc?")
+                   '((4 . "Branch matches superset of a previous branch"))))
+    (should (equal (xr-lint "abc\\|abcd*e?")
+                   '((5 . "Branch matches superset of a previous branch"))))
+    (should (equal (xr-lint "[a[:digit:]]\\|[a\n]")
+                   nil))
+    (should (equal (xr-lint "[a[:ascii:]]\\|[a\n]")
+                   '((14 . "Branch matches subset of a previous branch"))))
+
+    (should (equal (xr-lint "[[:alnum:]]\\|[[:alpha:]]")
+                   '((13 . "Branch matches subset of a previous branch"))))
+    (should (equal (xr-lint "[[:alnum:]%]\\|[[:alpha:]%]")
+                   '((14 . "Branch matches subset of a previous branch"))))
+    (should (equal (xr-lint "[[:xdigit:]%]\\|[[:alpha:]%]")
+                   nil))
+    (should (equal (xr-lint "[[:alnum:]]\\|[^[:alpha:]]")
+                   nil))
+    (should (equal (xr-lint "[^[:alnum:]]\\|[[:alpha:]]")
+                   nil))
+    (should (equal (xr-lint "[[:digit:]]\\|[^[:punct:]]")
+                   '((13 . "Branch matches superset of a previous branch"))))
+    (should (equal (xr-lint "[^[:digit:]]\\|[[:punct:]]")
+                   '((14 . "Branch matches subset of a previous branch"))))
+    (should (equal (xr-lint "[^[:digit:]]\\|[^[:xdigit:]]")
+                   '((14 . "Branch matches subset of a previous branch"))))
+    (should (equal (xr-lint "[^[:print:]]\\|[[:ascii:]]")
+                   nil))
+    (should (equal (xr-lint "[[:print:]]\\|[^[:ascii:]]")
+                   nil))
+    (should (equal (xr-lint "[^[:print:]]\\|[^[:ascii:]]")
+                   nil))
+    (should (equal (xr-lint "[[:digit:][:cntrl:]]\\|[[:ascii:]]")
+                   '((22 . "Branch matches superset of a previous branch"))))
+    (should (equal (xr-lint "[[:alpha:]]\\|A")
+                   '((13 . "Branch matches subset of a previous branch"))))
+    (should (equal (xr-lint "[[:alpha:]]\\|[A-E]")
+                   '((13 . "Branch matches subset of a previous branch"))))
+    (should (equal (xr-lint "[[:alpha:]3-7]\\|[A-E46]")
+                   '((16 . "Branch matches subset of a previous branch"))))
+    (should (equal (xr-lint "[^[:alpha:]]\\|[123]")
+                   '((14 . "Branch matches subset of a previous branch"))))
+    (should (equal (xr-lint "[!-@]\\|[[:digit:]]")
+                   '((7 . "Branch matches subset of a previous branch"))))
+    (should (equal (xr-lint "[^a-z]\\|[[:digit:]]")
+                   '((8 . "Branch matches subset of a previous branch"))))
+    (should (equal (xr-lint "[^[:punct:]]\\|[a-z]")
+                   '((14 . "Branch matches subset of a previous branch"))))
+    (should (equal (xr-lint "[[:space:]]\\|[ \t\f]")
+                   '((13 . "Branch matches subset of a previous branch"))))
+    (should (equal (xr-lint "[[:word:]]\\|[a-gH-P2357]")
+                   '((12 . "Branch matches subset of a previous branch"))))
+    (should (equal (xr-lint "[^[:space:]]\\|[a-gH-P2357]")
+                   '((14 . "Branch matches subset of a previous branch"))))
+    (should (equal (xr-lint "[^z-a]\\|[^0-9[:space:]]")
+                   '((8 . "Branch matches subset of a previous branch"))))
+
+    (should (equal (xr-lint "\\(?:.\\|\n\\)\\|a")
+                   '((12 . "Branch matches subset of a previous branch"))))
+    (should (equal (xr-lint "\\s-\\| ")
+                   '((5 . "Branch matches subset of a previous branch"))))
+    (should (equal (xr-lint "\\S-\\|x")
+                   '((5 . "Branch matches subset of a previous branch"))))
+    (should (equal (xr-lint "\\cl\\|å")
+                   '((5 . "Branch matches subset of a previous branch"))))
+    (should (equal (xr-lint "\\Ca\\|ü")
+                   '((5 . "Branch matches subset of a previous branch"))))
+    
+  ))
+
+(ert-deftest xr-lint-subsumed-repetition ()
+  (let ((text-quoting-style 'grave))
+    (should (equal (xr-lint "\\(?:a.c\\|def\\)+\\(?:abc\\)*")
+                   '((24 . "Repetition subsumed by preceding repetition"))))
+
+    ;; Exhaustive test of all possible combinations.
+    (should (equal (xr-lint "[ab]+a?,a?[ab]+,[ab]+a*,a*[ab]+")
+                   '((6 . "Repetition subsumed by preceding repetition")
+                     (14 . "Repetition subsumes preceding repetition")
+                     (22 . "Repetition subsumed by preceding repetition")
+                     (30 . "Repetition subsumes preceding repetition"))))
+
+    (should (equal (xr-lint "[ab]*a?,a?[ab]*,[ab]*a*,a*[ab]*")
+                   '((6 . "Repetition subsumed by preceding repetition")
+                     (14 . "Repetition subsumes preceding repetition")
+                     (22 . "Repetition subsumed by preceding repetition")
+                     (30 . "Repetition subsumes preceding repetition"))))
+
+    (should (equal (xr-lint "a+a?,a?a+,a+a*,a*a+,a*a?,a?a*,a*a*")
+                   '((3 . "Repetition subsumed by preceding repetition")
+                     (8 . "Repetition subsumes preceding repetition")
+                     (13 . "Repetition subsumed by preceding repetition")
+                     (18 . "Repetition subsumes preceding repetition")
+                     (23 . "Repetition subsumed by preceding repetition")
+                     (28 . "Repetition subsumes preceding repetition")
+                     (33 . "Repetition subsumed by preceding repetition"))))
+
+    (should (equal (xr-lint "[ab]+a??,a??[ab]+,[ab]+a*?,a*?[ab]+")
+                   '((6 . "Repetition subsumed by preceding repetition")
+                     (16 . "Repetition subsumes preceding repetition")
+                     (24 . "Repetition subsumed by preceding repetition")
+                     (34 . "Repetition subsumes preceding repetition"))))
+
+    (should (equal (xr-lint "[ab]*a??,a??[ab]*,[ab]*a*?,a*?[ab]*")
+                   '((6 . "Repetition subsumed by preceding repetition")
+                     (16 . "Repetition subsumes preceding repetition")
+                     (24 . "Repetition subsumed by preceding repetition")
+                     (34 . "Repetition subsumes preceding repetition"))))
+
+    (should (equal (xr-lint "a+a??,a??a+,a+a*?,a*?a+,a*a??,a??a*,a*a*?,a*?a*")
+                   '((3 . "Repetition subsumed by preceding repetition")
+                     (10 . "Repetition subsumes preceding repetition")
+                     (15 . "Repetition subsumed by preceding repetition")
+                     (22 . "Repetition subsumes preceding repetition")
+                     (27 . "Repetition subsumed by preceding repetition")
+                     (34 . "Repetition subsumes preceding repetition")
+                     (39 . "Repetition subsumed by preceding repetition")
+                     (46 . "Repetition subsumes preceding repetition"))))
+
+    (should (equal (xr-lint "[ab]+?a?,a?[ab]+?,[ab]+?a*,a*[ab]+?")
+                   nil))
+
+    (should (equal (xr-lint "[ab]*?a?,a?[ab]*?,[ab]*?a*,a*[ab]*?")
+                   nil))
+
+    (should (equal (xr-lint "a+?a?,a?a+?,a+?a*,a*a+?,a*?a?,a?a*?,a*?a*,a*a*?")
+                   '((40 . "Repetition subsumes preceding repetition")
+                     (45 . "Repetition subsumed by preceding repetition"))))
+
+    (should (equal (xr-lint "[ab]+?a??,a??[ab]+?,[ab]+?a*?,a*?[ab]+?")
+                   '((7 . "Repetition subsumed by preceding repetition")
+                     (17 . "Repetition subsumes preceding repetition")
+                     (27 . "Repetition subsumed by preceding repetition")
+                     (37 . "Repetition subsumes preceding repetition"))))
+
+    (should (equal (xr-lint "[ab]*?a??,a??[ab]*?,[ab]*?a*?,a*?[ab]*?")
+                   '((7 . "Repetition subsumed by preceding repetition")
+                     (17 . "Repetition subsumes preceding repetition")
+                     (27 . "Repetition subsumed by preceding repetition")
+                     (37 . "Repetition subsumes preceding repetition"))))
+
+    (should (equal (xr-lint "a+?a??,a??a+?,a+?a*?,a*?a+?,a*?a??,a??a*?,a*?a*?")
+                   '((4 . "Repetition subsumed by preceding repetition")
+                     (11 . "Repetition subsumes preceding repetition")
+                     (18 . "Repetition subsumed by preceding repetition")
+                     (25 . "Repetition subsumes preceding repetition")
+                     (32 . "Repetition subsumed by preceding repetition")
+                     (39 . "Repetition subsumes preceding repetition")
+                     (46 . "Repetition subsumed by preceding repetition"))))
+    ))
 
 (ert-deftest xr-skip-set ()
   (should (equal (xr-skip-set "0-9a-fA-F+*")
@@ -490,41 +630,42 @@
   )
 
 (ert-deftest xr-skip-set-lint ()
-  (should (equal (xr-skip-set-lint "A[:ascii:]B[:space:][:ascii:]")
-                 '((20 . "Duplicated character class `[:ascii:]'"))))
-  (should (equal (xr-skip-set-lint "a\\bF-AM-M\\")
-                 '((1 . "Unnecessarily escaped `b'")
-                   (3 . "Reversed range `F-A'")
-                   (6 . "Single-element range `M-M'")
-                   (9 . "Stray `\\' at end of string"))))
-  (should (equal (xr-skip-set-lint "A-Fa-z3D-KM-N!3-7\\!b")
-                 '((7 . "Ranges `A-F' and `D-K' overlap")
-                   (10 . "Two-element range `M-N'")
-                   (14 . "Range `3-7' includes character `3'")
-                   (17 . "Duplicated character `!'")
-                   (17 . "Unnecessarily escaped `!'")
-                   (19 . "Character `b' included in range `a-z'"))))
-  (should (equal (xr-skip-set-lint "!-\\$")
-                 '((2 . "Unnecessarily escaped `$'"))))
-  (should (equal (xr-skip-set-lint "[^a-z]")
-                 '((0 . "Suspect skip set framed in `[...]'"))))
-  (should (equal (xr-skip-set-lint "[0-9]+")
-                 '((0 . "Suspect skip set framed in `[...]'"))))
-  (should (equal (xr-skip-set-lint "[[:space:]].")
-                 '((0 . "Suspect character class framed in `[...]'"))))
-  (should (equal (xr-skip-set-lint "")
-                 '((0 . "Empty set matches nothing"))))
-  (should (equal (xr-skip-set-lint "^")
-                 '((0 . "Negated empty set matches anything"))))
-  (should (equal (xr-skip-set-lint "A-Z-")
-                 nil))
-  (should (equal (xr-skip-set-lint "-A-Z")
-                 nil))
-  (should (equal (xr-skip-set-lint "^-A-Z")
-                 nil))
-  (should (equal (xr-skip-set-lint "A-Z-z")
-                 '((3 . "Literal `-' not first or last"))))
-)
+  (let ((text-quoting-style 'grave))
+    (should (equal (xr-skip-set-lint "A[:ascii:]B[:space:][:ascii:]")
+                   '((20 . "Duplicated character class `[:ascii:]'"))))
+    (should (equal (xr-skip-set-lint "a\\bF-AM-M\\")
+                   '((1 . "Unnecessarily escaped `b'")
+                     (3 . "Reversed range `F-A'")
+                     (6 . "Single-element range `M-M'")
+                     (9 . "Stray `\\' at end of string"))))
+    (should (equal (xr-skip-set-lint "A-Fa-z3D-KM-N!3-7\\!b")
+                   '((7 . "Ranges `A-F' and `D-K' overlap")
+                     (10 . "Two-element range `M-N'")
+                     (14 . "Range `3-7' includes character `3'")
+                     (17 . "Duplicated character `!'")
+                     (17 . "Unnecessarily escaped `!'")
+                     (19 . "Character `b' included in range `a-z'"))))
+    (should (equal (xr-skip-set-lint "!-\\$")
+                   '((2 . "Unnecessarily escaped `$'"))))
+    (should (equal (xr-skip-set-lint "[^a-z]")
+                   '((0 . "Suspect skip set framed in `[...]'"))))
+    (should (equal (xr-skip-set-lint "[0-9]+")
+                   '((0 . "Suspect skip set framed in `[...]'"))))
+    (should (equal (xr-skip-set-lint "[[:space:]].")
+                   '((0 . "Suspect character class framed in `[...]'"))))
+    (should (equal (xr-skip-set-lint "")
+                   '((0 . "Empty set matches nothing"))))
+    (should (equal (xr-skip-set-lint "^")
+                   '((0 . "Negated empty set matches anything"))))
+    (should (equal (xr-skip-set-lint "A-Z-")
+                   nil))
+    (should (equal (xr-skip-set-lint "-A-Z")
+                   nil))
+    (should (equal (xr-skip-set-lint "^-A-Z")
+                   nil))
+    (should (equal (xr-skip-set-lint "A-Z-z")
+                   '((3 . "Literal `-' not first or last"))))
+    ))
 
 (provide 'xr-test)
 
