@@ -1,9 +1,9 @@
 ;;; xr.el --- Convert string regexp to rx notation   -*- lexical-binding: t -*-
 
-;; Copyright (C) 2019 Free Software Foundation, Inc.
+;; Copyright (C) 2019-2020 Free Software Foundation, Inc.
 
 ;; Author: Mattias Engdegård <mattiase@acm.org>
-;; Version: 1.15
+;; Version: 1.18
 ;; Package-Requires: ((emacs "26.1"))
 ;; URL: https://github.com/mattiase/xr
 ;; Keywords: lisp, regexps
@@ -23,64 +23,21 @@
 
 ;;; Commentary:
 
-;; This is an inverse companion to the rx package for translating
-;; regexps in string form to the rx notation.  Its chief uses are:
-;;
-;; - Migrating existing code to rx form, for better readability and
-;;   maintainability
-;; - Understanding complex regexp strings and finding errors in them
-;;   
-;; Please refer to `rx' for more information about the notation.
-;;
-;; In addition to Emacs regexps, this package can also parse and
-;; find mistakes in skip set strings, which are arguments to
-;; `skip-chars-forward' and `skip-chars-backward'.
-;;
-;; The exported functions are:
-;;
-;;  Regexps:
-;;  `xr'               - returns the converted rx expression
-;;  `xr-pp'            - converts to rx and pretty-prints
-;;  `xr-lint'          - finds mistakes in a regexp string
-;;
-;;  Skip sets:
-;;  `xr-skip-set'      - return the converted rx expression
-;;  `xr-skip-set-pp'   - converts to rx and pretty-prints
-;;  `xr-skip-set-lint' - finds mistakes in a skip set string
-;;
-;;  General:
-;;  `xr-pp-rx-to-str'  - pretty-prints an rx expression to a string
-;;
-;; Example (regexp found in compile.el):
-;;
-;;   (xr-pp "\\`\\(?:[^^]\\|\\^\\(?: \\*\\|\\[\\)\\)")
-;; =>
-;;   (seq bos
-;;        (or (not (any "^"))
-;;            (seq "^"
-;;                 (or " *" "["))))
-;;
-;; The rx notation admits many synonyms. The user is encouraged to
-;; edit the result for maximum readability, consistency and personal
-;; preference when replacing existing regexps in elisp code.
-;;
-;; Related work:
-;;
-;; The `lex' package, a lexical analyser generator, provides the
-;; `lex-parse-re' function which translates regexps to rx, but does
-;; not attempt to handle all the edge cases of Elisp's regexp syntax
-;; or pretty-print the result.
-;;
-;; The `pcre2el' package, a regexp syntax converter and interactive
-;; regexp explainer, could also be used for translating regexps to rx.
-;; `xr' is narrower in scope but more accurate for the purpose of
-;; parsing Emacs regexps and printing the results in rx form.
-;;
-;; Neither of these packages parse skip-set strings or provide
-;; mistake-finding functions.
+;; This package translates regexps in string form to the rx notation.
+;; It can also find mistakes and questionable constructs in regexps
+;; and related expressions. See the README file for more information.
 
 ;;; News:
 
+;; Version 1.18:
+;; - Fix test broken in Emacs 26
+;; Version 1.17:
+;; - Performance improvements
+;; Version 1.16:
+;; - Translate [^\n] into nonl
+;; - Better character class subset/superset analysis
+;; - More accurate repetition subsumption check
+;; - Use text quoting for messages
 ;; Version 1.15:
 ;; - Warn about subsuming repetitions in sequence, like [AB]+A*
 ;; Version 1.14:
@@ -139,12 +96,14 @@
         (if (>= end ?\])
             (push (vector ?\] end (point)) intervals)
           (xr--report warnings (point)
-                      (format "Reversed range `%s' matches nothing"
-                              (xr--escape-string (match-string 0) nil))))
+                      (format-message
+                       "Reversed range `%s' matches nothing"
+                       (xr--escape-string (match-string 0) nil))))
         (when (eq end ?^)
           (xr--report warnings (point)
-                      (format "Two-character range `%s'"
-                              (xr--escape-string (match-string 0) nil)))))
+                      (format-message
+                       "Two-character range `%s'"
+                       (xr--escape-string (match-string 0) nil)))))
       (goto-char (match-end 0)))
      ;; Initial ]
      ((looking-at "]")
@@ -154,7 +113,7 @@
     (while (not (looking-at "]"))
       (cond
        ;; character class
-       ((looking-at (rx "[:" (group (*? anything)) ":]"))
+       ((looking-at (rx "[:" (group (* (not (any ":")))) ":]"))
         (let ((sym (intern (match-string 1))))
           (unless (memq sym
                         '(ascii alnum alpha blank cntrl digit graph
@@ -163,13 +122,14 @@
             (error "No character class `%s'" (match-string 0)))
           (if (memq sym classes)
               (xr--report warnings (point)
-                          (format "Duplicated character class `[:%s:]'" sym))
+                          (format-message
+                           "Duplicated character class `[:%s:]'" sym))
             (push sym classes))
           (goto-char (match-end 0))))
        ;; character range
-       ((looking-at (rx (group (not (any "]"))) "-" (group (not (any "]")))))
-        (let ((start (string-to-char (match-string 1)))
-              (end   (string-to-char (match-string 2))))
+       ((looking-at (rx (not (any "]")) "-" (not (any "]"))))
+        (let ((start (char-after))
+              (end   (char-after (+ (point) 2))))
           (cond
            ((<= start end)
             (push (vector start end (point)) intervals))
@@ -177,14 +137,16 @@
            ((and (eq start ?z) (eq end ?a)))
            (t
             (xr--report warnings (point)
-                        (format "Reversed range `%s' matches nothing"
-                                (xr--escape-string (match-string 0) nil)))))
+                        (format-message
+                         "Reversed range `%s' matches nothing"
+                         (xr--escape-string (match-string 0) nil)))))
           ;; Suppress warnings about ranges between adjacent digits,
           ;; like [0-1], as they are common and harmless.
           (when (and (= end (1+ start)) (not (<= ?0 start end ?9)))
             (xr--report warnings (point)
-                        (format "Two-character range `%s'"
-                                (xr--escape-string (match-string 0) nil))))
+                        (format-message
+                         "Two-character range `%s'"
+                         (xr--escape-string (match-string 0) nil))))
           (goto-char (match-end 0))))
        ((looking-at (rx eos))
         (error "Unterminated character alternative"))
@@ -205,12 +167,13 @@
                      (not (and intervals
                                (eq (aref (car (last intervals)) 0) ?\]))))
             (xr--report warnings (point)
-                        "Suspect `[' in char alternative"))
+                        (format-message "Suspect `[' in char alternative")))
           (when (and (looking-at (rx "-" (not (any "]"))))
                      (> (point) start-pos))
             (xr--report
              warnings (point)
-             "Literal `-' not first or last in character alternative"))
+             (format-message
+              "Literal `-' not first or last in character alternative")))
           (push (vector ch ch (point)) intervals))
         (forward-char 1))))
 
@@ -231,33 +194,37 @@
                     ((and (eq (aref this 0) (aref this 1))
                           (eq (aref next 0) (aref next 1)))
                      (setcdr s (cddr s))
-                     (format "Duplicated `%c' inside character alternative"
-                             (aref this 0)))
+                     (format-message
+                      "Duplicated `%c' inside character alternative"
+                      (aref this 0)))
                     ;; Duplicate range: drop it and warn.
                     ((and (eq (aref this 0) (aref next 0))
                           (eq (aref this 1) (aref next 1)))
                      (setcdr s (cddr s))
-                     (format "Duplicated `%c-%c' inside character alternative"
-                             (aref this 0) (aref this 1)))
+                     (format-message
+                      "Duplicated `%c-%c' inside character alternative"
+                      (aref this 0) (aref this 1)))
                     ;; Character in range: drop it and warn.
                     ((eq (aref this 0) (aref this 1))
                      (setcar s next)
                      (setcdr s (cddr s))
-                     (format "Character `%c' included in range `%c-%c'"
-                             (aref this 0) (aref next 0) (aref next 1)))
+                     (format-message
+                      "Character `%c' included in range `%c-%c'"
+                      (aref this 0) (aref next 0) (aref next 1)))
                     ;; Same but other way around.
                     ((eq (aref next 0) (aref next 1))
                      (setcdr s (cddr s))
-                     (format "Character `%c' included in range `%c-%c'"
-                             (aref next 0) (aref this 0) (aref this 1)))
+                     (format-message
+                      "Character `%c' included in range `%c-%c'"
+                      (aref next 0) (aref this 0) (aref this 1)))
                     ;; Overlapping ranges: merge and warn.
                     (t
                      (let ((this-end (aref this 1)))
                        (aset this 1 (max (aref this 1) (aref next 1)))
                        (setcdr s (cddr s))
-                       (format "Ranges `%c-%c' and `%c-%c' overlap"
-                               (aref this 0) this-end
-                               (aref next 0) (aref next 1)))))))
+                       (format-message "Ranges `%c-%c' and `%c-%c' overlap"
+                                       (aref this 0) this-end
+                                       (aref next 0) (aref next 1)))))))
               (xr--report warnings (max (aref this 2) (aref next 2))
                           (xr--escape-string message nil)))))
         (setq s (cdr s)))
@@ -297,13 +264,19 @@
           (if negated
               (list 'not (car classes))
             (car classes)))
+         ;; [^\n]: nonl.
+         ((and negated
+               (equal chars '(?\n))
+               (null ranges)
+               (null classes))
+          'nonl)
          ;; Anything else: produce (any ...)
          (t
           ;; Put dash last of all single characters.
           (when (memq ?- chars)
             (setq chars (cons ?- (delq ?- chars))))
           (let* ((set (cons 'any
-                            (append
+                            (nconc
                              (and ranges
                                   (list (apply #'concat (nreverse ranges))))
                              (and chars
@@ -314,14 +287,14 @@
               set))))))))
 
 (defun xr--rev-join-seq (sequence)
-  "Reverse a sequence, flatten any (seq ...) inside, and concatenate
-adjacent strings."
+  "Reverse SEQUENCE, flatten any (seq ...) inside, and concatenate
+adjacent strings. SEQUENCE is used destructively."
   (let ((result nil))
     (while sequence
       (let ((elem (car sequence))
             (rest (cdr sequence)))
         (cond ((and (consp elem) (eq (car elem) 'seq))
-               (setq sequence (append (reverse (cdr elem)) rest)))
+               (setq sequence (nconc (nreverse (cdr elem)) rest)))
               ((and (stringp elem) (stringp (car result)))
                (setq result (cons (concat elem (car result)) (cdr result)))
                (setq sequence rest))
@@ -476,7 +449,8 @@ UPPER may be nil, meaning infinity."
           (forward-char 1)
           (if (null sequence)
               (push 'bol sequence)
-            (xr--report warnings (match-beginning 0) "Unescaped literal `^'")
+            (xr--report warnings (match-beginning 0)
+                        (format-message "Unescaped literal `^'"))
             (push "^" sequence)))
 
          ;; $ - only special at end of sequence
@@ -484,7 +458,8 @@ UPPER may be nil, meaning infinity."
           (forward-char 1)
           (if (looking-at (rx (or "\\|" "\\)" eos)))
               (push 'eol sequence)
-            (xr--report warnings (match-beginning 0) "Unescaped literal `$'")
+            (xr--report warnings (match-beginning 0)
+                        (format-message "Unescaped literal `$'"))
             (push "$" sequence)))
 
          ;; * ? + (and non-greedy variants)
@@ -534,7 +509,7 @@ UPPER may be nil, meaning infinity."
             (let ((literal (match-string 1)))
               (goto-char (match-end 1))
               (xr--report warnings (match-beginning 0)
-                          (format "Unescaped literal `%s'" literal))
+                          (format-message "Unescaped literal `%s'" literal))
               (push literal sequence))))
 
          ;; \{..\} - not special at beginning of sequence or after ^
@@ -601,7 +576,7 @@ UPPER may be nil, meaning infinity."
          ;; character alternative
          ((looking-at (rx "[" (opt (group "^"))))
           (goto-char (match-end 0))
-          (let ((negated (match-string 1)))
+          (let ((negated (match-beginning 1)))
             (push (xr--parse-char-alt negated warnings) sequence)))
 
          ;; group
@@ -609,9 +584,9 @@ UPPER may be nil, meaning infinity."
                                      (opt (opt (group (any "1-9")
                                                       (zero-or-more digit)))
                                           (group ":")))))
-          (let ((question (match-string 1))
+          (let ((question (match-beginning 1))
                 (number (match-string 2))
-                (colon (match-string 3)))
+                (colon (match-beginning 3)))
             (when (and question (not colon))
               (error "Invalid \\(? syntax"))
             (goto-char (match-end 0))
@@ -686,14 +661,14 @@ UPPER may be nil, meaning infinity."
          ;; accept any not otherwise handled character after the backslash
          ;; since such sequences are found in the wild.
          ((looking-at (rx "\\" (group (or (any "\\*+?.^$[]")
-                                          (group anything)))))
+                                          (group (not (any "\\*+?.^$[]")))))))
           (forward-char 2)
           (push (match-string 1) sequence)
           (when (match-beginning 2)
             ;; Note that we do not warn about \\], since the symmetry with \\[
             ;; makes it unlikely to be a serious error.
             (xr--report warnings (match-beginning 0)
-                        (format "Escaped non-special character `%s'"
+                        (format-message "Escaped non-special character `%s'"
                                 (xr--escape-string (match-string 2) nil)))))
 
          (t (error "Backslash at end of regexp")))
@@ -701,9 +676,13 @@ UPPER may be nil, meaning infinity."
         (when (and warnings (cdr sequence))
           ;; Check for subsuming repetitions in sequence: (Rx X) (Ry Y)
           ;; where Rx and Ry are repetition operators, and X and Y are operands.
-          ;; We conclude that (Rx X) subsumes (Ry Y) if Rx can match
-          ;; infinitely many times, Ry can match zero times,
-          ;; and X matches a superset of Y. Example: [ab]+a?
+          ;; We conclude that (Rx X) subsumes (Ry Y), in the sense that the
+          ;; sequence is equivalent to just (Rx X), if:
+          ;;       X matches a superset of Y
+          ;;   and Rx can match infinitely many times
+          ;;   and Ry can match zero times
+          ;;   and Ry is non-greedy if Rx is non-greedy.
+          ;; Example: [ab]+a?
           (let* ((item (car sequence))
                  (expr (and (consp item)
                             (memq (car item)
@@ -717,19 +696,26 @@ UPPER may be nil, meaning infinity."
                                  '(zero-or-more one-or-more opt *? +? ??))
                            (xr--make-seq (cdr prev-item)))))
                 (when prev-expr
-                  (cond
-                   ((and (memq (car item) '(zero-or-more opt *? ??))
-                         (memq (car prev-item)
-                               '(zero-or-more one-or-more *? +?))
-                         (xr--superset-p prev-expr expr))
-                    (xr--report warnings item-start
-                                "Repetition subsumed by preceding repetition"))
-                   ((and (memq (car prev-item) '(zero-or-more opt *? ??))
-                         (memq (car item) '(zero-or-more one-or-more *? +?))
-                         (xr--superset-p expr prev-expr))
-                    (xr--report
-                     warnings item-start
-                     "Repetition subsumes preceding repetition"))))))))))
+                  (let ((op (car item))
+                        (prev-op (car prev-item)))
+                    ;; Test the same condition twice, but mirrored.
+                    (cond
+                     ((and (memq op '(zero-or-more opt *? ??))
+                           (memq prev-op '(zero-or-more one-or-more *? +?))
+                           (not (and (memq prev-op '(*? +?))
+                                     (memq op '(zero-or-more opt))))
+                           (xr--superset-p prev-expr expr))
+                      (xr--report
+                       warnings item-start
+                       "Repetition subsumed by preceding repetition"))
+                     ((and (memq prev-op '(zero-or-more opt *? ??))
+                           (memq op '(zero-or-more one-or-more *? +?))
+                           (not (and (memq op '(*? +?))
+                                     (memq prev-op '(zero-or-more opt))))
+                           (xr--superset-p expr prev-expr))
+                      (xr--report
+                       warnings item-start
+                       "Repetition subsumes preceding repetition")))))))))))
 
     (let ((item-seq (xr--rev-join-seq sequence)))
       (cond ((null item-seq)
@@ -753,30 +739,121 @@ UPPER may be nil, meaning infinity."
 (defun xr--any-arg-to-items (arg)
   "Convert an `any' argument to a list of characters, ranges (as pairs),
 and classes (symbols)."
-  ;; We know (since we built it) that x is either a symbol or
-  ;; a string, and that the string does not mix ranges and chars.
-  (cond ((symbolp arg) (list arg))
+  ;; We know (since we built it) that x is either a symbol, string or char,
+  ;; and that the string does not mix ranges and chars.
+  (cond ((symbolp arg)
+         ;; unibyte and multibyte are aliases of ascii and nonascii in
+         ;; practice; simplify.
+         (list (cond ((eq arg 'unibyte) 'ascii)
+                     ((eq arg 'multibyte) 'nonascii)
+                     (t arg))))
+        ((characterp arg) (list arg))
         ((and (>= (length arg) 3)
               (eq (aref arg 1) ?-))
          (xr--range-string-to-items arg))
         (t (string-to-list arg))))
 
+;; Character class relation matrix
+;; Legend:  = same
+;;          ⊂ row subset of column
+;;          ⊃ row superset of column
+;;          x overlapping
+;;          ∅ disjoint
+;;          ? not certain but very likely
+;;          * assuming `case-fold-search' is nil
+;;
+;;         alp aln dig xdi cnt asc non bla gra pri pun upp low spa wor
+;; alpha    =   ⊂   ∅   x   ∅   x   x   ∅   ⊂   ⊂   ∅?  ⊃?  ⊃?  ∅?  ⊂?
+;; alnum    ⊃   =   ⊃   ⊃   ∅   x   x   ∅   ⊂   ⊂   ∅?  ⊃?  ⊃?  ∅?  ⊂?
+;; digit    ∅   ⊂   =   ⊂   ∅   ⊂   ∅   ∅   ⊂   ⊂   ∅   ∅?  ∅?  ∅?  ⊂?
+;; xdigit   x   ⊂   ⊃   =   ∅   ⊂   ∅   ∅   ⊂   ⊂   ∅   x?  x?  ∅?  ⊂?
+;; cntrl    ∅   ∅   ∅   ∅   =   ⊂   ∅   x   ∅   ∅   ∅   ∅?  ∅?  x?  ∅?
+;; ascii    x   x   ⊃   ⊃   ⊃   =   ∅   x   x   x   x   x?  x?  x?  x?
+;; nonascii x   x   ∅   ∅   ∅   ∅   =   x   x   x   x?  x?  x?  x?  x?
+;; blank    ∅   ∅   ∅   ∅   x   x   x   =   ∅   x   x?  ∅?  ∅?  x?  ∅?
+;; graph    ⊃   ⊃   ⊃   ⊃   ∅   x   x   ∅   =   ⊂   ⊃?  ⊃?  ⊃?  ∅?  ⊃?
+;; print    ⊃   ⊃   ⊃   ⊃   ∅   x   x   x   ⊃   =   ⊃?  ⊃?  ⊃?  x?  ⊃?
+;; punct    ∅?  ∅?  ∅   ∅   ∅   x   x?  x?  ⊂?  ⊂?  =   ∅?  ∅?  ∅?  x?
+;; upper    ⊂?  ⊂?  ∅?  x?  ∅?  x?  x?  ∅?  ⊂?  ⊂?  ∅?  =   ∅*  ∅?  ⊂?
+;; lower    ⊂?  ⊂?  ∅?  x?  ∅?  x?  x?  ∅?  ⊂?  ⊂?  ∅?  ∅*  =   ∅?  ⊂?
+;; space    ∅?  ∅?  ∅?  ∅?  x?  x?  x?  x?  ∅?  x?  ∅?  ∅?  ∅?  =   ∅
+;; word     ⊃?  ⊃?  ⊃?  ⊃?  ∅?  x?  x?  ∅?  ⊂?  ⊂?  x?  ⊃?  ⊃?  ∅   =
+
 (defun xr--any-item-superset-p (a b)
   "Whether A is a superset of B, both being `any' items: a character,
 a range (pair of chars), or a class (symbol)."
   (cond
    ((symbolp a)
-    (cond ((symbolp b) (eq a b))
-          ((eq b ?\n)
-           (memq a '(alnum alpha blank digit graph
-                     lower multibyte nonascii print punct space
-                     upper word xdigit)))))
+    (cond
+     ((symbolp b)
+      (or (eq a b)
+          (memq
+           b
+           (cdr (assq
+                 a
+                 ;; Class superset matrix: first class in each row is
+                 ;; a superset of all the rest in that row.
+                 ;; It is slightly approximative, since upper, lower
+                 ;; and (partially) punct can be modified through case
+                 ;; and syntax tables.
+                 '((alpha upper lower)
+                   (alnum alpha digit xdigit upper lower)
+                   (xdigit digit)
+                   (ascii digit xdigit cntrl)
+                   (graph alpha alnum digit xdigit punct upper lower word)
+                   (print alpha alnum digit xdigit graph punct
+                          upper lower word)
+                   (word alpha alnum digit xdigit upper lower)))))))
+
+     ((characterp b)
+      (cond
+       ;; Some reasonable subsets of `space' and `word'.
+       ((eq a 'space) (memq b '(?\s ?\t ?\f)))
+       ((eq a 'word)
+        (string-match-p (rx (any "0-9A-Za-z")) (char-to-string b)))
+       ;; Test for invariant classes only. `punct' is invariant for ASCII.
+       ;; `upper' and `lower' are not really invariant but mostly.
+       ((or (memq a '(digit xdigit cntrl ascii nonascii alpha alnum blank
+                            graph print upper lower))
+            (and (eq a 'punct) (<= b 127)))
+        (string-match-p (format "[[:%s:]]" a) (char-to-string b)))))
+
+     (t   ; b is a range.
+      ;; For simplicity, only check ASCII ranges.
+      (and (<= (cdr b) 127)
+           (cl-some
+            (lambda (a-range) (and (<= (car a-range) (car b))
+                                   (<= (cdr b) (cdr a-range))))
+            (cdr (assq a '((alpha (?A . ?Z) (?a . ?z))
+                           (alnum (?0 . ?9) (?A . ?Z) (?a . ?z))
+                           (digit (?0 . ?9))
+                           (xdigit (?0 . ?9) (?A . ?F) (?a . ?f))
+                           (cntrl (0 . 31))
+                           (ascii (0 . 127))
+                           (graph (33 . 126))
+                           (print (32 . 126))
+                           (punct (33 . 47) (58 . 64) (91 . 96) (123 . 126))
+                           ;; Not-so-wild assumptions.
+                           (upper (?A . ?Z))
+                           (lower (?a . ?z))
+                           (word (?0 . ?9) (?A . ?Z) (?a . ?z))
+                           (space (?\s . ?\s) (?\t . ?\t) (?\f . ?\f))))))))))
+   
    ((consp a)
-    (or (and (characterp b)
-             (<= (car a) b (cdr a)))
-        (and (consp b)
-             (<= (car a) (car b) (cdr b) (cdr a)))))
-   (t
+    (cond
+     ((characterp b) (<= (car a) b (cdr a)))
+     ((consp b) (<= (car a) (car b) (cdr b) (cdr a)))
+     (t   ; b is a class.
+      ;; Only consider classes with simple definitions.
+      (let ((b-hull (cdr (assq b '((digit . (?0 . ?9))
+                                   (xdigit . (?0 . ?f))
+                                   (cntrl . (0 . 31))
+                                   (ascii . (0 . 127))
+                                   (nonascii . (#x80 . #x10ffff)))))))
+        (and b-hull
+             (<= (car a) (car b-hull))
+             (<= (cdr b-hull) (cdr a)))))))
+   (t   ; a is a character.
     (and (characterp b) (eq a b)))))
 
 (defun xr--any-item-may-intersect-p (a b)
@@ -784,18 +861,89 @@ a range (pair of chars), or a class (symbol)."
 a range (pair of chars), or a class (symbol). If in doubt, return t."
   (cond
    ((symbolp a)
-    (cond ((eq b ?\n)
-           (not (memq a '(alnum alpha blank digit graph
-                          lower multibyte nonascii print punct space
-                          upper word xdigit))))
-          (t t)))
+    (cond
+     ((symbolp b)
+      (or (eq a b)
+          (memq
+           b
+           (cdr (assq
+                 a
+                 ;; Class intersection matrix: first class in each row
+                 ;; intersects all the rest in that row.
+                 ;; Again slightly approximate, since upper, lower,
+                 ;; space, word and (partially) punct can be modified
+                 ;; through syntax and case tables.
+                 '((alpha alnum xdigit ascii nonascii graph print
+                          upper lower word)
+                   (alnum alpha digit xdigit ascii nonascii graph print
+                          upper lower word)
+                   (digit alnum xdigit ascii graph print word)
+                   (xdigit alpha alnum digit ascii graph print
+                           upper lower word)
+                   (cntrl ascii blank space)
+                   (ascii alpha alnum digit xdigit cntrl ascii blank
+                          graph print punct upper lower space word)
+                   (nonascii alpha alnum blank graph print punct
+                             upper lower space word)
+                   (blank cntrl ascii nonascii print punct space)
+                   (graph alpha alnum digit xdigit ascii nonascii print punct
+                          upper lower word)
+                   (print alpha alnum digit xdigit ascii nonascii blank graph
+                          punct upper lower space word)
+                   (punct ascii nonascii blank graph print upper lower word)
+                   (upper alpha alnum xdigit ascii nonascii graph print word)
+                   (lower alpha alnum xdigit ascii nonascii graph print word)
+                   (space cntrl ascii nonascii blank print)
+                   (word alpha alnum digit xdigit ascii nonascii graph print
+                         punct upper lower)))))))
+
+     ((characterp b)
+      (cond
+       ;; Some reasonably conservative subsets of `space' and `word'.
+       ((eq a 'space)
+        (not (string-match-p (rx (any (33 . 126))) (char-to-string b))))
+       ((eq a 'word)
+        (not (memq b '(?\s ?\t ?\f ?\r))))
+       (t
+        ;; Only some classes are invariant. `punct' is invariant for ASCII.
+        ;; `upper' and `lower' are not really invariant but mostly.
+        (or (and (eq a 'punct) (> b 127))
+            ;; This may be a tad slow.
+            (string-match-p (format "[[:%s:]]" a) (char-to-string b))))))
+
+     (t   ; b is a range.
+      ;; For simplicity, only check ASCII ranges.
+      (cond
+       ((and (> (cdr b) 127)
+             (not (memq a '(cntrl ascii digit xdigit)))))
+       ((eq a 'space)
+        (not (cl-some (lambda (a-range) (and (<= (car a-range) (cdr b))
+                                             (<= (car b) (cdr a-range))))
+                      '((?0 . ?9) (?A . ?Z) (?a . ?z)))))
+       ((eq a 'word))
+       (t
+        (cl-some
+         (lambda (a-range) (and (<= (car a-range) (cdr b))
+                                (<= (car b) (cdr a-range))))
+         (cdr (assq a '((alpha (?A . ?Z) (?a . ?z))
+                        (alnum (?0 . ?9) (?A . ?Z) (?a . ?z))
+                        (digit (?0 . ?9))
+                        (xdigit (?0 . ?9) (?A . ?F) (?a . ?f))
+                        (cntrl (0 . 31))
+                        (ascii (0 . 127))
+                        (graph (33 . 126))
+                        (print (32 . 126))
+                        (punct (33 . 47) (58 . 64) (91 . 96) (123 . 126))
+                        ;; Not-so-wild assumptions.
+                        (upper (?A . ?Z))
+                        (lower (?a . ?z)))))))))))
+
    ((consp a)
-    (or (and (characterp b)
-             (<= (car a) b (cdr a)))
-        (and (consp b)
-             (<= (car a) (cdr b))
-             (<= (car b) (cdr a)))
-        (symbolp b)))
+    (cond ((characterp b) (<= (car a) b (cdr a)))
+          ((consp b) (and (<= (car a) (cdr b))
+                          (<= (car b) (cdr a))))
+          (t  ; b is a class
+           (xr--any-item-may-intersect-p b a))))
    ;; Now a must be a character.
    ((characterp b) (eq a b))
    (t (xr--any-item-may-intersect-p b a))))
@@ -827,10 +975,15 @@ A-SETS and B-SETS are arguments to `any'."
     ((or 'ascii 'alnum 'alpha 'blank 'cntrl 'digit 'graph
          'lower 'multibyte 'nonascii 'print 'punct 'space
          'unibyte 'upper 'word 'xdigit)
-     (xr--char-superset-of-char-set-p sets negated `(any ,rx)))
-    ((pred stringp)
-     (and (= (length rx) 1)
-          (xr--char-superset-of-char-set-p sets negated (list rx))))))
+     (xr--char-superset-of-char-set-p sets negated (list rx)))
+    (`(not ,(and sym
+                 (or 'ascii 'alnum 'alpha 'blank 'cntrl 'digit 'graph
+                     'lower 'multibyte 'nonascii 'print 'punct 'space
+                     'unibyte 'upper 'word 'xdigit)))
+     (and negated
+          (xr--char-superset-of-char-set-p (list sym) nil sets)))
+    ((pred characterp)
+     (xr--char-superset-of-char-set-p sets negated (list rx)))))
 
 (defun xr--single-non-newline-char-p (rx)
   "Whether RX only matches single characters none of which is newline."
@@ -840,6 +993,18 @@ A-SETS and B-SETS are arguments to `any'."
     (`(syntax ,s) (not (eq s ?>)))      ; comment-end often matches newline
     (_ (xr--char-superset-of-rx-p '("\n") t rx))))
 
+(defun xr--single-char-p (rx)
+  "Whether RX only matches single characters."
+  (or (memq rx '(nonl anything
+                 ascii alnum alpha blank cntrl digit graph
+                 lower multibyte nonascii print punct space
+                 unibyte upper word xdigit))
+      (characterp rx)
+      (and (consp rx)
+           (or (memq (car rx) '(any category syntax))
+               (and (eq (car rx) 'not)
+                    (xr--single-char-p (cadr rx)))))))
+
 (defun xr--syntax-superset-of-rx-p (syntax negated rx)
   "Whether SYNTAX, possibly NEGATED, is a superset of RX."
   ;; Syntax tables vary, but we make a (quite conservative) guess.
@@ -859,24 +1024,21 @@ A-SETS and B-SETS are arguments to `any'."
     (and set
          (xr--char-superset-of-rx-p (cdr set) nil rx))))
 
-(defun xr--string-to-chars (str)
-  (mapcar #'char-to-string (string-to-list str)))
-
 (defun xr--expand-strings (rx)
-  "If RX is a string or a seq of strings, convert them to seqs of
-single-character strings."
+  "Expand strings to characters or seqs of characters.
+`seq' forms are expanded non-recursively."
   (cond ((consp rx)
          (if (eq (car rx) 'seq)
              (cons 'seq (mapcan (lambda (x)
-                                  (if (and (stringp x)
-                                           (> (length x) 1))
-                                      (xr--string-to-chars x)
+                                  (if (stringp x)
+                                      (string-to-list x)
                                     (list x)))
                                 (cdr rx)))
            rx))
-        ((and (stringp rx)
-              (> (length rx) 1))
-         (cons 'seq (xr--string-to-chars rx)))
+        ((stringp rx)
+         (if (= (length rx) 1)
+             (string-to-char rx)
+           (cons 'seq (string-to-list rx))))
         (t rx)))
 
 (defun xr--superset-seq-p (a b)
@@ -907,8 +1069,20 @@ single-character strings."
         (xr--char-superset-of-rx-p sets nil b))
        (`(not (any . ,sets))
         (xr--char-superset-of-rx-p sets t b))
+       ((or 'ascii 'alnum 'alpha 'blank 'cntrl 'digit 'graph
+            'lower 'multibyte 'nonascii 'print 'punct 'space
+            'unibyte 'upper 'word 'xdigit)
+        (xr--char-superset-of-rx-p (list a) nil b))
+       (`(not ,(and sym
+                    (or 'ascii 'alnum 'alpha 'blank 'cntrl 'digit 'graph
+                        'lower 'multibyte 'nonascii 'print 'punct 'space
+                        'unibyte 'upper 'word 'xdigit)))
+        (xr--char-superset-of-rx-p (list sym) t b))
+
        ('nonl (xr--single-non-newline-char-p b))
 
+       ('anything (xr--single-char-p b))
+
        (`(seq . ,a-body)
         (pcase b
           (`(seq . ,b-body)
@@ -952,8 +1126,8 @@ single-character strings."
 
        ((or `(category ,_) `(not (category ,_)))
         (or (equal a b)
-            (and (stringp b)
-                 (string-match-p (rx-to-string a) b))))
+            (and (characterp b)
+                 (string-match-p (rx-to-string a) (char-to-string b)))))
 
        (_ (equal a b))))))
 
@@ -979,10 +1153,9 @@ single-character strings."
         (push seq alternatives)))
     (if (cdr alternatives)
         ;; Simplify (or nonl "\n") to anything
-        (if (or (equal alternatives '(nonl "\n"))
-                (equal alternatives '("\n" nonl)))
+        (if (member alternatives '((nonl "\n") ("\n" nonl)))
             'anything
-          (cons 'or (reverse alternatives)))
+          (cons 'or (nreverse alternatives)))
       (car alternatives))))
 
 (defun xr--parse (re-string warnings)
@@ -1016,7 +1189,8 @@ single-character strings."
                                   (opt "?"))
                              eos))
              (not (looking-at (rx "[:" (one-or-more anything) ":]" eos))))
-    (xr--report warnings (point) "Suspect skip set framed in `[...]'"))
+    (xr--report warnings (point)
+                (format-message "Suspect skip set framed in `[...]'")))
 
   (let ((negated (looking-at (rx "^")))
         (start-pos (point))
@@ -1038,11 +1212,12 @@ single-character strings."
           (when (and (eq (char-before) ?\[)
                      (eq (char-after (match-end 0)) ?\]))
             (xr--report warnings (1- (point))
-                        "Suspect character class framed in `[...]'"))
+                        (format-message
+                         "Suspect character class framed in `[...]'")))
           (when (memq sym classes)
             (xr--report warnings (point)
-                        (format "Duplicated character class `%s'"
-                                (match-string 0))))
+                        (format-message "Duplicated character class `%s'"
+                                        (match-string 0))))
           (push sym classes)))
 
        ((looking-at (rx (or (seq "\\" (group anything))
@@ -1060,32 +1235,37 @@ single-character strings."
                      (not (memq start '(?^ ?- ?\\))))
             (xr--report warnings (point)
                         (xr--escape-string
-                         (format "Unnecessarily escaped `%c'" start) nil)))
+                         (format-message "Unnecessarily escaped `%c'" start)
+                         nil)))
           (when (and (match-beginning 3)
                      (not (memq end '(?^ ?- ?\\))))
             (xr--report warnings (1- (match-beginning 3))
                         (xr--escape-string
-                         (format "Unnecessarily escaped `%c'" end) nil)))
+                         (format-message "Unnecessarily escaped `%c'" end)
+                         nil)))
           (when (and (eq start ?-)
                      (not end)
                      (match-beginning 2)
                      (< start-pos (point) (1- (point-max))))
             (xr--report warnings (point)
-                        "Literal `-' not first or last"))
+                        (format-message "Literal `-' not first or last")))
           (if (and end (> start end))
               (xr--report warnings (point)
                           (xr--escape-string
-                           (format "Reversed range `%c-%c'" start end) nil))
+                           (format-message "Reversed range `%c-%c'" start end)
+                           nil))
             (cond
              ((eq start end)
               (xr--report warnings (point)
                           (xr--escape-string
-                           (format "Single-element range `%c-%c'" start end)
+                           (format-message "Single-element range `%c-%c'"
+                                           start end)
                            nil)))
              ((eq (1+ start) end)
               (xr--report warnings (point)
                           (xr--escape-string
-                           (format "Two-element range `%c-%c'" start end)
+                           (format-message "Two-element range `%c-%c'"
+                                           start end)
                            nil))))
             (let ((tail ranges))
               (while tail
@@ -1096,18 +1276,20 @@ single-character strings."
                              (cond
                               ((and end (< start end)
                                     (< (car range) (cdr range)))
-                               (format "Ranges `%c-%c' and `%c-%c' overlap"
-                                       (car range) (cdr range) start end))
+                               (format-message
+                                "Ranges `%c-%c' and `%c-%c' overlap"
+                                (car range) (cdr range) start end))
                               ((and end (< start end))
-                               (format "Range `%c-%c' includes character `%c'"
-                                       start end (car range)))
+                               (format-message
+                                "Range `%c-%c' includes character `%c'"
+                                start end (car range)))
                               ((< (car range) (cdr range))
-                               (format
+                               (format-message
                                 "Character `%c' included in range `%c-%c'"
                                 start (car range) (cdr range)))
                               (t
-                               (format "Duplicated character `%c'"
-                                       start)))))
+                               (format-message "Duplicated character `%c'"
+                                               start)))))
                         (xr--report warnings (point)
                                     (xr--escape-string msg nil))
                         ;; Expand previous interval to include this range.
@@ -1121,7 +1303,7 @@ single-character strings."
 
        ((looking-at (rx "\\" eos))
         (xr--report warnings (point)
-                    "Stray `\\' at end of string")))
+                    (format-message "Stray `\\' at end of string"))))
 
       (goto-char (match-end 0)))
 
